/**
 * This file is part of the DOM implementation for KDE.
 *
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2003 Apple Computer, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include "Text.h"

#include "Document.h"
#include "ExceptionCode.h"
#include "RenderText.h"
#include "TextBreakIterator.h"

#if PLATFORM(AG)
#include "FontData.h"
#include "FontPlatformDataAG.h"
#include "RenderArena.h"
#include "FontCache.h"
#include "settings.h"
#include "document.h"
#include "frame.h"
#include "cssstyleselector.h"
#include "WebKitAGPriv.h"
#endif

#if ENABLE(SVG)
#include "RenderSVGInlineText.h"
#endif // ENABLE(SVG)

namespace WebCore {

// DOM Section 1.1.1

// ### allow having children in text nodes for entities, comments etc.

Text::Text(Document *doc, const String &_text)
    : CharacterData(doc, _text)
{
}

Text::Text(Document *doc)
    : CharacterData(doc)
{
}

Text::~Text()
{
}

Text *Text::splitText(unsigned offset, ExceptionCode& ec)
{
    ec = 0;

    // FIXME: This does not copy markers
    
    // INDEX_SIZE_ERR: Raised if the specified offset is negative or greater than
    // the number of 16-bit units in data.
    if (offset > str->length()) {
        ec = INDEX_SIZE_ERR;
        return 0;
    }

    // NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
    if (isReadOnlyNode()) {
        ec = NO_MODIFICATION_ALLOWED_ERR;
        return 0;
    }

    StringImpl *oldStr = str;
    Text *newText = createNew(str->substring(offset, str->length()-offset));
    str = str->copy();
    str->ref();
    str->remove(offset, str->length()-offset);

    dispatchModifiedEvent(oldStr);
    oldStr->deref();

    if (parentNode())
        parentNode()->insertBefore(newText,nextSibling(), ec);
    if (ec)
        return 0;

    if (renderer())
        static_cast<RenderText*>(renderer())->setText(str);

    return newText;
}

const AtomicString& Text::localName() const
{
    return textAtom;
}

String Text::nodeName() const
{
    return textAtom.domString();
}

Node::NodeType Text::nodeType() const
{
    return TEXT_NODE;
}

PassRefPtr<Node> Text::cloneNode(bool /*deep*/)
{
    return document()->createTextNode(str);
}

bool Text::rendererIsNeeded(RenderStyle *style)
{
    if (!CharacterData::rendererIsNeeded(style))
        return false;

    bool onlyWS = containsOnlyWhitespace();
    if (!onlyWS)
        return true;

    RenderObject *par = parentNode()->renderer();
    
    if (par->isTable() || par->isTableRow() || par->isTableSection() || par->isTableCol() || par->isFrameSet())
        return false;
    
    if (style->preserveNewline()) // pre/pre-wrap/pre-line always make renderers.
        return true;
    
    RenderObject *prev = previousRenderer();
    if (prev && prev->isBR()) // <span><br/> <br/></span>
        return false;
        
    if (par->isInlineFlow()) {
        // <span><div/> <div/></span>
        if (prev && !prev->isInline())
            return false;
    } else {
        if (par->isRenderBlock() && !par->childrenInline() && (!prev || !prev->isInline()))
            return false;
        
        RenderObject *first = par->firstChild();
        while (first && first->isFloatingOrPositioned())
            first = first->nextSibling();
        RenderObject *next = nextRenderer();
        if (!first || next == first)
            // Whitespace at the start of a block just goes away.  Don't even
            // make a render object for this text.
            return false;
    }
    
    return true;
}

RenderObject *Text::createRenderer(RenderArena *arena, RenderStyle *style)
{
#if ENABLE(SVG)
    if (parentNode()->isSVGElement())
        return new (arena) RenderSVGInlineText(this, str);
#endif // ENABLE(SVG)
    
    return new (arena) RenderText(this, str);
}

void Text::attach()
{
    createRendererIfNeeded();
    CharacterData::attach();
}

void Text::recalcStyle( StyleChange change )
{
    if (change != NoChange && parentNode())
        if (renderer())
            renderer()->setStyle(parentNode()->renderer()->style());
    if (changed() && renderer() && renderer()->isText())
        static_cast<RenderText*>(renderer())->setText(str);
    setChanged(NoStyleChange);
}

// DOM Section 1.1.1
bool Text::childTypeAllowed(NodeType)
{
    return false;
}

Text *Text::createNew(StringImpl *_str)
{
    return new Text(document(), _str);
}

String Text::toString() const
{
    // FIXME: substitute entity references as needed!
    return nodeValue();
}

PassRefPtr<Text> Text::createWithLengthLimit(Document* doc, const String& text, unsigned& charsLeft, unsigned maxChars)
{
    if (charsLeft == text.length() && charsLeft <= maxChars) {
        charsLeft = 0;
        return new Text(doc, text);
    }
    
    unsigned start = text.length() - charsLeft;
    unsigned end = start + std::min(charsLeft, maxChars);
    
    // check we are not on an unbreakable boundary
    TextBreakIterator* it = characterBreakIterator(text.characters(), text.length());
    if (end < text.length() && !isTextBreak(it, end))
        end = textBreakPreceding(it, end);
        
    // maxChars of unbreakable characters could lead to infinite loop
    if (end <= start)
        end = text.length();
    
    String nodeText = text.substring(start, end - start);
    charsLeft = text.length() - end;
        
    return new Text(doc, nodeText);
}

#ifndef NDEBUG
void Text::formatForDebugger(char *buffer, unsigned length) const
{
    String result;
    String s;
    
    s = nodeName();
    if (s.length() > 0) {
        result += s;
    }
          
    s = nodeValue();
    if (s.length() > 0) {
        if (result.length() > 0)
            result += "; ";
        result += "value=";
        result += s;
    }
          
    strncpy(buffer, result.deprecatedString().latin1(), length - 1);
}
#endif

#if PLATFORM(AG)

static bool isCollapsibleWhiteSpace(UChar c, bool preserveNewLine, bool collapseWhiteSpace) {
        switch (c) {
            case ' ':
			case '\t':
                return collapseWhiteSpace;
            case '\n':
                return !preserveNewLine;
        }
        return false;
}

static bool isZeroWidthSpace(UChar c) { return c < 0x20 || (c >= 0x7F && c < 0xA0) || c == 0x200e || c == 0x200f; }

static bool shouldReplaceWithSpace(UChar c) 
{
	if(c == 0x1680 || c == 0x180E ||	(c >= 0x2002 && c <= 0x200D) ||
		c == 0x202F || c == 0x205F || c == 0x2060 || c == 0x3000 || c == 0xFEFF ) 
		return true;
	
	return false;
}

void Text::stripNonPrintableCharacters(bool preserveNewLine, bool collapseWhiteSpace)
{
	unsigned start = 0;
    unsigned end = str->length();
	UChar nextChar;
	UChar space = ' ';
	int lastNewLine = 0;
    
    bool collapsing = false;
	while (start < end)
	{
		nextChar = str->characters()[start];
		
		if(shouldReplaceWithSpace(nextChar))
		{
			str->remove(start,1);
			str->insert(&space, 1, start);
			nextChar = space;
		}

		if(isCollapsibleWhiteSpace(nextChar, preserveNewLine, collapseWhiteSpace))
		{
			if(collapsing)
			{
				--end;
				str->remove(start,1);				
			}
			else
			{
				collapsing = true;
				str->remove(start,1);
				str->insert(&space, 1, start);
				++start;
			}
		}
		else if(isZeroWidthSpace(nextChar) && nextChar != '\t' && nextChar != '\n')
		{
			collapsing = false;
			--end;
			str->remove(start,1);
		}
		else
		{
			collapsing = false;
			if(nextChar == '\n')
				lastNewLine = start+1;

			if(nextChar == '\t')
			{
				str->remove(start,1);
				int spaceCount = 8 - (start - lastNewLine)%8;

				for(int i=0; i < spaceCount ; i++)
					str->insert(&space, 1, start);
				
				start += (spaceCount-1);
				end += (spaceCount - 1);
			}
			++start;
		}
	}
}

typedef struct _TextRunFont
{
	int startRange;
	int endRange;
	FontDescription fd;
} TextRunFont;

const UChar* Text::textContentWebKitAGNode(unsigned int* l)
{
	if (l) 
	{
		*l = 0;
	}
	if (this && renderer() && nodeType() == TEXT_NODE)
	{
		RenderText* rt = static_cast<RenderText*>(renderer());
		if (rt)
		{
			String s = rt->RenderText::text();
			if (l)
			{
				*l = s.length();
			}
			return s.characters();
		}
	}
	return 0;
}

bool Text::splitNodeForCorrectFont()
{
	RenderObject* ro = renderer();

	setHasBeenCheckedForFont(true);

	bool bWasSplitNode = false;

	Vector<TextRunFont> textRuns;
	if (ro)
	{
		RenderText* rt = static_cast<RenderText*>(ro);
		if (rt)
		{
			RenderStyle* rs = rt->style();
			if (! rs)
			{
				return bWasSplitNode;
			}
			// Both FontDescription and Font should be got by reference
			// so we do not make unnecessary copies.
			const FontDescription& fd = rs->fontDescription();
			const Font& f = rs->font();
			const FontData* fontData = f.primaryFont();
			StringImpl* rtext = rt->RenderText::text();
			const UChar* s = rtext ? rtext->characters() : 0;
			unsigned textLen = rtext ? rtext->length() : 0;

			const AtomicString curFontFamily = fd.firstFamily().family();
			const FontPlatformData& fontPlatformData = fontData->platformData();
			AGFontBackend_t* fontBE = 0;
			if (fontPlatformData.fontServerBackend())
				fontBE = fontPlatformData.fontServerBackend()->ag_font_backend;

			bool bSplittingRequired = false;
			if (fontPlatformData.scaledFont() == 0)
				bSplittingRequired = true;

			if (fontBE)
			{
				for(unsigned j = 0; j < textLen; ++j) {
					// WEBC: TODO: Handle surrogates correctly.
					if (f.glyphDataForCharacter(s[j], s+j, 1, false, false).glyph == 0) {
						bSplittingRequired = true;
						break;
					}
				}
			}
			if (bSplittingRequired)
			{
				bWasSplitNode = true;

				Document* doc = document();
				ASSERT(doc);
				WebKitAGLanguageScript langScript = doc->guessedLanguageScript();

				AGScaledFont_t* agOrigScaledFont = fontPlatformData.scaledFont();
				TextRunFont curTextRun;

				curTextRun.startRange = 0;
				curTextRun.endRange = 0;
				curTextRun.fd = fd;

				FontDescription prevFD = fd;

				for (unsigned i = 0; i < textLen; ++i)
				{
					FontDescription thisFD = fd;

					AGStatus_t foundInOrigFont = AG_STATUS_SUCCESS;
					if (agOrigScaledFont)
					{
						// WEBC: TODO: Handle surrogates correctly.
						foundInOrigFont = (f.glyphDataForCharacter(s[i], s+i, 1, false, false).glyph == 0) ? 
							AG_STATUS_FOUND_UNSUPPORTED_CHAR : AG_STATUS_SUCCESS;
					}
					else
					{
						foundInOrigFont = AG_STATUS_FOUND_UNSUPPORTED_CHAR;
					}

					if (foundInOrigFont == AG_STATUS_FOUND_UNSUPPORTED_CHAR)
					{
						//check our fallback font
						const FontDescription& oldFD = rs->fontDescription();
						//createFontDescriptionForScript(fontBE, newFD, oldFD, langScript);
						AGFont_t* agFont = fontBE->ag_get_font_from_scaled_font(fontBE, agOrigScaledFont);
						createFontDescriptionForChar(fontBE, thisFD, oldFD, s[i], agFont, langScript);
					}
					else
					{
						thisFD = fd;
					}

					if (thisFD != prevFD)
					{
						if (i != 0)
						{
							textRuns.append(curTextRun);
						}
						curTextRun.fd = thisFD;
						curTextRun.startRange = curTextRun.endRange = i;
						prevFD = thisFD;
					}
					else
					{
						curTextRun.endRange = i;
					}
				}
				textRuns.append(curTextRun);
			}
			if (! textRuns.isEmpty())
			{
				ExceptionCode ec;
				for (Vector<TextRunFont>::iterator itr = textRuns.begin(); itr != textRuns.end(); ++itr)
				{
					PassRefPtr<Node> clonedNode = cloneNode(true);
					if (clonedNode.get())
					{
						TextRunFont tr = *itr;
						int sI = tr.startRange, eI = tr.endRange;
						String clonedNodeTextContent(s+sI, eI-sI+1);
						clonedNode->setTextContent(clonedNodeTextContent, ec);
						Node* n = static_cast<Node*>(this);
						if (parentNode())
							parentNode()->insertBefore(clonedNode.get(), n, ec);
						clonedNode->setHasBeenCheckedForFont(true);
						//clonedNode->createRendererIfNeeded();
						Text* clonedTextNode = static_cast<Text*>(clonedNode.get());
						RenderObject* clonedNodeRO = clonedTextNode->renderer();
						if (clonedNodeRO)
						{
							RenderText* clonedNodeRT = static_cast<RenderText*>(clonedNodeRO);
							RenderStyle* oldRenderStyle = clonedNodeRT->style();
							RenderStyle* rs = ::new (document()->renderArena()->allocate(sizeof(RenderStyle)))RenderStyle(*oldRenderStyle);
							if (rs)
							{
								rs->setFontDescription(tr.fd);
								rs->font().update();
							}
							if (clonedNodeRT)
								clonedNodeRT->setStyle(rs);
						}
					}
				}
				String blankString;
				setTextContent(blankString, ec);
				RenderObject* toRemoveRO = renderer();
				if (toRemoveRO)
				{
					RenderObject* p = toRemoveRO->parent();
					if (p)
					{
						p->removeChildNode(toRemoveRO);
					}
					static_cast<RenderText*>(toRemoveRO)->setText(blankString.impl());
					toRemoveRO->destroy();
				}
				setRenderer(0);
				/*RenderText* origRenderText = static_cast<RenderText*>(origRenderObj);
				if (origRenderText)
				{
				origRenderText->setText(s.impl(), true);
				}*/
			}
		}
	}
	return bWasSplitNode;
}

void Text::createFontDescriptionForChar(AGFontBackend_t* fontBE, FontDescription& newFD, const FontDescription& oldFD, UChar32 c, AGFont_t* agOrigFont, WebKitAGLanguageScript langScript)
{
	newFD.setUsePrinterFont(oldFD.usePrinterFont());
	const Settings* settings = 0;
	
	if (document() && document()->frame())
		settings = document()->frame()->settings();
	
	const WebKitAGFontLanguageScriptMapEntry* tagUnicodeFontMap = settings->GetFontForLanguage(AG_LANG_SCRIPT_UNICODE);
	const AtomicString sUnicodeFont = tagUnicodeFontMap->fontName;
	
	const WebKitAGFontLanguageScriptMapEntry* tagLanguageFontMap = settings->GetFontForLanguage(langScript);
	const AtomicString sLangFont = tagLanguageFontMap->fontName;

	AGEncodingType_t bestEncodingType = AG_ENCODING_PLATFORM;
	
	AGEncodingType_t encodingType = AG_ENCODING_PLATFORM;
#if PLATFORM(WIN)
	encodingType = AG_ENCODING_WIN_PLATFORM;
#elif PLATFORM(MAC)
	encodingType = AG_ENCODING_MAC_PLATFORM;
#endif

	const char* bestFontForChar = fontBE->ag_get_best_font_and_encoding_for_character(c, agOrigFont, langScript, &bestEncodingType);
	AtomicString sFont = bestFontForChar;
	if (bestFontForChar)
		WebKitAG::g_HostFunctions->freeBytes((void*)bestFontForChar);

	if (sFont == nullAtom)
		sFont = sLangFont;
	if (sFont == nullAtom)
	{
		sFont = sUnicodeFont;
	}
	const AtomicString newFontName = sFont;

	if (!newFontName.isEmpty()) 
	{
		newFD.firstFamily().setFamily(newFontName);
		newFD.firstFamily().appendFamily(0);
		newFD.setSpecifiedSize(oldFD.specifiedSize());
		newFD.setComputedSize(oldFD.computedSize());
	}
	AGEncoding_t* encoding = fontBE->ag_create_encoding(fontBE, bestEncodingType);
	newFD.setEncoding(encoding);
}

void Text::createFontDescriptionForScript(AGFontBackend_t* fontBE, FontDescription& newFD, const FontDescription& oldFD, WebKitAGLanguageScript langScript)
{
	newFD.setUsePrinterFont(oldFD.usePrinterFont());
	const Settings* settings = 0;
	
	if (document() && document()->frame())
		settings = document()->frame()->settings();
	
	const WebKitAGFontLanguageScriptMapEntry* tagUnicodeFontMap = settings->GetFontForLanguage(AG_LANG_SCRIPT_UNICODE);
	const AtomicString sUnicodeFont = tagUnicodeFontMap->fontName;
	
	const WebKitAGFontLanguageScriptMapEntry* tagLanguageFontMap = settings->GetFontForLanguage(langScript);
	const AtomicString sLangFont = tagLanguageFontMap->fontName;

	AtomicString sFont = sLangFont;
	if(sFont == nullAtom)
	{
		langScript = AG_LANG_SCRIPT_UNICODE;
		sFont = sUnicodeFont;
	}
	const AtomicString newFontName = sFont;

	if (!newFontName.isEmpty()) 
	{
		newFD.firstFamily().setFamily(newFontName);
		newFD.firstFamily().appendFamily(0);
		newFD.setSpecifiedSize(oldFD.specifiedSize());
		newFD.setComputedSize(oldFD.computedSize());
	}
	AGEncoding_t* encoding = 0;
	switch (langScript)
	{
	case AG_LANG_SCRIPT_UNICODE:
		encoding = fontBE->ag_create_encoding(fontBE, AG_ENCODING_UNICODE_BE);
		break;
	case AG_LANG_SCRIPT_WESTERN:
		encoding = fontBE->ag_create_encoding(fontBE, AG_ENCODING_WIN_PLATFORM);
		break;
	default:
		{
			const char* cmap_identityH = "Identity-H";
			encoding = fontBE->ag_create_encoding_from_cmap_name(fontBE, cmap_identityH, strlen(cmap_identityH));
		}
	}
	newFD.setEncoding(encoding);
}
#endif

} // namespace WebCore
