#include "config.h"
#include "FontCache.h"

#include "Font.h"
#include "FontFallbackList.h"
#include "FontPlatformDataAG.h"
#include "StringHash.h"
#include <wtf/HashMap.h>

namespace WebCore {

struct FontPlatformDataCacheKey {
    FontPlatformDataCacheKey(const AtomicString& family = AtomicString(), unsigned size = 0, bool bold = false, bool italic = false, AGEncodingType_t encodingType = AG_ENCODING_DEFAULT)
        : m_family(family)
        , m_size(size)
        , m_bold(bold)
        , m_italic(italic)
		, m_encodingType(encodingType)
    {
    }

    bool operator==(const FontPlatformDataCacheKey& other) const
    {
        return equalIgnoringCase(m_family, other.m_family) && m_size == other.m_size && m_bold == other.m_bold && m_italic == other.m_italic && m_encodingType == other.m_encodingType;
    }
    
    AtomicString m_family;
    unsigned m_size;
    bool m_bold;
    bool m_italic;	
	AGEncodingType_t m_encodingType;
};

inline unsigned computeHash(const FontPlatformDataCacheKey& fontKey)
{
    unsigned hashCodes[4] = {
        CaseInsensitiveHash<String>::hash(fontKey.m_family),
        fontKey.m_size,
        static_cast<unsigned>(fontKey.m_bold) << 1 | static_cast<unsigned>(fontKey.m_italic),
		fontKey.m_encodingType
    };
    return StringImpl::computeHash(reinterpret_cast<UChar*>(hashCodes), 4 * sizeof(unsigned) / sizeof(UChar));
}

struct FontPlatformDataCacheKeyHash {
    static unsigned hash(const FontPlatformDataCacheKey& font)
    {
        return computeHash(font);
    }
         
    static bool equal(const FontPlatformDataCacheKey& a, const FontPlatformDataCacheKey& b)
    {
        return a == b;
    }
};

struct FontPlatformDataCacheKeyTraits : WTF::GenericHashTraits<FontPlatformDataCacheKey> {
    static const bool emptyValueIsZero = true;
    static const bool needsDestruction = false;
    static const FontPlatformDataCacheKey& deletedValue()
    {
        static FontPlatformDataCacheKey key(nullAtom, 0xFFFFFFFFU, false, false);
        return key;
    }
    static const FontPlatformDataCacheKey& emptyValue()
    {
        static FontPlatformDataCacheKey key(nullAtom, 0, false, false);
        return key;
    }
};

typedef HashMap<FontPlatformDataCacheKey, FontPlatformData*, FontPlatformDataCacheKeyHash, FontPlatformDataCacheKeyTraits> FontPlatformDataCache;

static FontPlatformDataCache* gFontPlatformDataCache = 0;

static const AtomicString& alternateFamilyName(const AtomicString& familyName)
{
    // Alias Courier <-> Courier New
    static AtomicString courier("Courier"), courierNew("Courier New");
    if (equalIgnoringCase(familyName, courier))
        return courierNew;
    if (equalIgnoringCase(familyName, courierNew))
        return courier;

    // Alias Times and Times New Roman.
    static AtomicString times("Times"), timesNewRoman("Times New Roman");
    if (equalIgnoringCase(familyName, times))
        return timesNewRoman;
    if (equalIgnoringCase(familyName, timesNewRoman))
        return times;
    
	// Alias Times-Roman -> Times New Roman.
    static AtomicString timesRoman("Times-Roman");
    if (equalIgnoringCase(familyName, timesRoman))
        return timesNewRoman;

    // Alias Arial and Helvetica
    static AtomicString arial("Arial"), helvetica("Helvetica");
    if (equalIgnoringCase(familyName, arial))
        return helvetica;
    if (equalIgnoringCase(familyName, helvetica))
        return arial;

    return emptyAtom;
}

FontPlatformData* FontCache::getCachedFontPlatformData(const FontDescription& fontDescription, const AtomicString& familyName, bool checkingAlternateName)
{
    if (!gFontPlatformDataCache) {
        gFontPlatformDataCache = new FontPlatformDataCache;
        platformInit();
    }

    FontPlatformDataCacheKey key(familyName
		, fontDescription.computedPixelSize()
		, fontDescription.bold()
		, fontDescription.italic()
		, fontDescription.encoding() ? fontDescription.encoding()->type : AG_ENCODING_DEFAULT
		);

    FontPlatformData* result = 0;
    bool foundResult = false;
    FontPlatformDataCache::iterator it = gFontPlatformDataCache->find(key);
    if (it == gFontPlatformDataCache->end()) {
        result = createFontPlatformData(fontDescription, familyName);
		if (result)
		{
			gFontPlatformDataCache->set(key, result);
			foundResult = true;
		}
    } else {
        result = it->second;
        foundResult = true;
    }

    if (!foundResult && !checkingAlternateName) {
        // We were unable to find a font.  We have a small set of fonts that we alias to other names, 
        // e.g., Arial/Helvetica, Courier/Courier New, etc.  Try looking up the font under the aliased name.
        const AtomicString& alternateName = alternateFamilyName(familyName);
        if (!alternateName.isEmpty())
            result = getCachedFontPlatformData(fontDescription, alternateName, true);
		if (result) {
			FontPlatformData * platformData = new FontPlatformData(*result);
			platformData->setFontObjectsNeedDestruction(false); // This is an aliased instance of the "core" ag font objects, and does not need releasing
			gFontPlatformDataCache->set(key, platformData); // Cache the result under the old name.
		}
    }

    return result;
}

struct FontDataCacheKeyHash {
    static unsigned hash(const FontPlatformData& platformData)
    {
        return platformData.hash();
    }
         
    static bool equal(const FontPlatformData& a, const FontPlatformData& b)
    {
        return a == b;
    }
};

struct FontDataCacheKeyTraits : WTF::GenericHashTraits<FontPlatformData> {
    static const bool emptyValueIsZero = true;
	// AG Impl's FontPlatformData has refcounted objects that need correct destruction
    static const bool needsDestruction = true;
    static const FontPlatformData& deletedValue()
    {
        static FontPlatformData key = FontPlatformData::Deleted();
        return key;
    }
    static const FontPlatformData& emptyValue()
    {
        static FontPlatformData key;
        return key;
    }
};

typedef HashMap<FontPlatformData, FontData*, FontDataCacheKeyHash, FontDataCacheKeyTraits> FontDataCache;

static FontDataCache* gFontDataCache = 0;

FontData* FontCache::getCachedFontData(const FontPlatformData* platformData)
{
    if (!platformData)
        return 0;

    if (!gFontDataCache)
        gFontDataCache = new FontDataCache;
    
    FontData* result = gFontDataCache->get(*platformData);
    if (!result) {
        result = new FontData(*platformData);
        gFontDataCache->set(*platformData, result);
    }
        
    return result;
}

const FontData* FontCache::getFontData(const Font& font, int& familyIndex)
{
    FontPlatformData* result = 0;

    int startIndex = familyIndex;
    const FontFamily* startFamily = &font.fontDescription().family();
    for (int i = 0; startFamily && i < startIndex; i++)
        startFamily = startFamily->next();
    const FontFamily* currFamily = startFamily;
    while (currFamily && !result) {
        familyIndex++;
        if (currFamily->family().length())
            result = getCachedFontPlatformData(font.fontDescription(), currFamily->family());
        currFamily = currFamily->next();
    }

    if (!currFamily)
        familyIndex = cAllFamiliesScanned;

    if (!result)
        // We didn't find a font. Try to find a similar font using our own specific knowledge about our platform.
        // For example on OS X, we know to map any families containing the words Arabic, Pashto, or Urdu to the
        // Geeza Pro font.
        result = getSimilarFontPlatformData(font);

    if (!result && startIndex == 0)
        // We still don't have a result.  Hand back our last resort fallback font.  We only do the last resort fallback
        // when trying to find the primary font.  Otherwise our fallback will rely on the actual characters used.
        result = getLastResortFallbackFont(font);

    // Now that we have a result, we need to go from FontPlatformData -> FontData.
    return getCachedFontData(result);
}

void FontCache::platformInit()
{
	// Not needed on Windows.
}

#if PLATFORM(WIN)
IMLangFontLink2* FontCache::getFontLinkInterface()
{
	static IMultiLanguage *multiLanguage;
	if (!multiLanguage) {
		if (CoCreateInstance(CLSID_CMultiLanguage, 0, CLSCTX_ALL, IID_IMultiLanguage, (void**)&multiLanguage) != S_OK)
			return 0;
	}

	static IMLangFontLink2* langFontLink;
	if (!langFontLink) {
		if (multiLanguage->QueryInterface(IID_IMLangFontLink2, (void**)&langFontLink) != S_OK)
			return 0;
	}

	return langFontLink;
}
#endif

FontPlatformData* FontCache::getSimilarFontPlatformData(const Font& font)
{
	return 0;
}

FontPlatformData* FontCache::getLastResortFallbackFont(const Font& font)
{
	// FIXME: Would be even better to somehow get the user's default font here.  For now we'll pick
	// the default that the user would get without changing any prefs.
	static AtomicString timesStr("Times New Roman");
	return getCachedFontPlatformData(font.fontDescription(), timesStr);
}

FontPlatformData* FontCache::createFontPlatformData(const FontDescription& fontDescription, const AtomicString& family)
{
	int fontNameLen = family.length();
	const UChar* fontName = family.characters();
	FontPlatformData* pFontPlatformData = new FontPlatformData(fontName, fontNameLen, fontDescription.computedPixelSize(), (unsigned int)fontDescription.weight(), fontDescription.italic(), fontDescription.encoding());
	if (pFontPlatformData)
	{
		if (pFontPlatformData->font() == 0 || pFontPlatformData->scaledFont() == 0)
		{
			delete pFontPlatformData;
			return 0;
		}
	}
	return pFontPlatformData;
}

const FontData* FontCache::getFontDataForCharacters(const Font& font, const UChar* characters, int length)
{
	// Default to Tahoma
	FontPlatformData* result = getCachedFontPlatformData(font.fontDescription(), "Tahoma", true);
    if (result)
		return getCachedFontData(result);
    return NULL;
}

// Clears the font cache
void FontCache::purgeFontCache() {
	if(gFontPlatformDataCache) {
		for(FontPlatformDataCache::iterator j = gFontPlatformDataCache->begin(); j != gFontPlatformDataCache->end(); ++j) {
			FontPlatformData* pData = (*j).second;
			pData->releaseObjects();
			delete pData;
		}
		gFontPlatformDataCache->clear();
	}
	if(gFontDataCache) {
		for(FontDataCache::iterator j = gFontDataCache->begin(); j != gFontDataCache->end(); ++j) {
			FontData * pData = (*j).second;
			delete pData;
		}
		gFontDataCache->clear();
	}
}


} // namespace WebCore
