/*************************************************************************
*
* ADOBE CONFIDENTIAL
* ___________________
*
*  Copyright 2007 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains
* the property of Adobe Systems Incorporated and its suppliers,
* if any.  The intellectual and technical concepts contained
* herein are proprietary to Adobe Systems Incorporated and its
* suppliers and may be covered by U.S. and Foreign Patents,
* patents in process, and are protected by trade secret or copyright law.
* Dissemination of this information or reproduction of this material
* is strictly forbidden unless prior written permission is obtained
* from Adobe Systems Incorporated.
**************************************************************************/
//this include is necessary to build on mac.
//otherwise we get this compiler error: if you use new/delete make sure to include config.h at the top of the file
#include "config.h"

#if PLATFORM(AG)
#include "Image.h"
#include "FloatRect.h"
#include "GraphicsContext.h"
#include <math.h>
#include "ag.h"
#include "AffineTransform.h"
#include "PlatformString.h"
#include "BitmapImage.h"

PassRefPtr<WebCore::SharedBuffer> loadResourceIntoBuffer(const char*);

namespace WebCore {

	void BitmapImage::initPlatformData()
	{
		setPlatformData(NULL);
	}

	void BitmapImage::invalidatePlatformData()
	{
	}

	
	Image* Image::loadPlatformResource(const char *name)
	{
		RefPtr<SharedBuffer> buffer = loadResourceIntoBuffer(name);
		BitmapImage* img = new BitmapImage;
		img->setData(buffer.release(), true);
		return img;
	}

#if PLATFORM(WIN)
	bool BitmapImage::getHBITMAP(HBITMAP bmp)
	{
		return false;
	}
#endif	
	
	bool Image::supportsType(const String& type)
	{
		/* We support the types that are handled in createDecoder() */
		if(type == "image/png"
			|| type == "image/x-png"
			|| type == "image/jpeg"
			|| type == "image/jpg"
			|| type == "image/gif"
			|| type == "image/xbm"
			|| type == "image/x-xbm"
			|| type == "image/bmp")
		{
			return true;
		} else {
			return false;
		}
	}

	void FrameData::clear()
	{
		if (m_frame) 
		{
			AGImage_t* image = m_frame;
			if (image)
			{
				if (image->rgbaData)
					free(image->rgbaData);
				free(image);
			}
			m_frame = 0;
			m_duration = 0.;
		}
	}

	
	// ================================================
	// Image Class
	// ================================================

	// Drawing Routines
	void BitmapImage::checkForSolidColor()
	{
		agNotImplemented();
		m_isSolidColor = false;
	}

	static inline FloatSize calculatePatternScale(const FloatRect& dstRect, const FloatRect& srcRect, Image::TileRule hRule, Image::TileRule vRule)
	{
		float scaleX = 1.0f, scaleY = 1.0f;

		if (hRule == Image::StretchTile)
			scaleX = dstRect.width() / srcRect.width();
		if (vRule == Image::StretchTile)
			scaleY = dstRect.height() / srcRect.height();

		if (hRule == Image::RepeatTile)
			scaleX = scaleY;
		if (vRule == Image::RepeatTile)
			scaleY = scaleX;

		return FloatSize(scaleX, scaleY);
	}

	void Image::drawTiled(GraphicsContext* ctxt, const FloatRect& destRect, const FloatPoint& srcPoint, const FloatSize& scaledTileSize, CompositeOperator op)
	{    
		if (!nativeImageForCurrentFrame())
			return;

		if (mayFillWithSolidColor()) {
			fillWithSolidColor(ctxt, destRect, solidColor(), op);
			return;
		}

		FloatSize intrinsicTileSize = size();
		FloatSize scale(scaledTileSize.width() / intrinsicTileSize.width(),
			scaledTileSize.height() / intrinsicTileSize.height());
		
		AffineTransform patternTransform = AffineTransform().scale(scale.width(), scale.height());

		FloatRect oneTileRect;
		oneTileRect.setX(destRect.x() + fmodf(fmodf(-srcPoint.x(), scaledTileSize.width()) - scaledTileSize.width(), scaledTileSize.width()));
		oneTileRect.setY(destRect.y() + fmodf(fmodf(-srcPoint.y(), scaledTileSize.height()) - scaledTileSize.height(), scaledTileSize.height()));
		oneTileRect.setSize(scaledTileSize);

		// Check and see if a single draw of the image can cover the entire area we are supposed to tile.    
		if (oneTileRect.contains(destRect)) {
			FloatRect visibleSrcRect;
			visibleSrcRect.setX((destRect.x() - oneTileRect.x()) / scale.width());
			visibleSrcRect.setY((destRect.y() - oneTileRect.y()) / scale.height());
			visibleSrcRect.setWidth(destRect.width() / scale.width());
			visibleSrcRect.setHeight(destRect.height() / scale.height());
			draw(ctxt, destRect, visibleSrcRect, op);
			return;
		}
		else
		{
			oneTileRect.setX((destRect.x() - oneTileRect.x()) / scale.width());
			oneTileRect.setY((destRect.y() - oneTileRect.y()) / scale.height());
		}

		FloatRect tileRect(FloatPoint(), intrinsicTileSize);    
		drawPattern(ctxt, tileRect, patternTransform, oneTileRect.location(), op, destRect);
	}

// FIXME: Merge with the other drawTiled eventually, since we need a combination of both for some things.
	void Image::drawTiled(GraphicsContext* ctxt, const FloatRect& dstRect, const FloatRect& srcRect, TileRule hRule, TileRule vRule, CompositeOperator op)
	{    
		if (!nativeImageForCurrentFrame())
			return;

		if (mayFillWithSolidColor()) {
			fillWithSolidColor(ctxt, dstRect, solidColor(), op);
			return;
		}

		// FIXME: We do not support 'round' yet.  For now just map it to 'repeat'.
		if (hRule == RoundTile)
			hRule = RepeatTile;
		if (vRule == RoundTile)
			vRule = RepeatTile;

		FloatSize scale = calculatePatternScale(dstRect, srcRect, hRule, vRule);
		AffineTransform patternTransform = AffineTransform().scale(scale.width(), scale.height());

		// We want to construct the phase such that the pattern is centered (when stretch is not
		// set for a particular rule).
		float hPhase = scale.width() * srcRect.x();
		float vPhase = scale.height() * srcRect.y();
		if (hRule == Image::RepeatTile)
			hPhase -= fmodf(dstRect.width(), scale.width() * srcRect.width()) / 2.0f;
		if (vRule == Image::RepeatTile)
			vPhase -= fmodf(dstRect.height(), scale.height() * srcRect.height()) / 2.0f;
		FloatPoint patternPhase(dstRect.x() - hPhase, dstRect.y() - vPhase);

		drawPattern(ctxt, srcRect, patternTransform, patternPhase, op, dstRect);

		//startAnimation();
	}

	void Image::drawPattern(GraphicsContext* ctxt, const FloatRect& tileRect, const AffineTransform& patternTransform,
						const FloatPoint& phase, CompositeOperator op, const FloatRect& destRect)
	{
		if (!ctxt || !ctxt->platformContext())
			return;

#if HANDLE_PAGE_BREAKS
		if (ctxt->isDrawnOnLastPage(destRect.y(), destRect.height()))
			return;
#endif

		AGContext_t* context = ctxt->platformContext();
		context->ag_backend->ag_save(context);

		AGImage_t* tileImage = nativeImageForCurrentFrame();
		float w = tileImage->imageWidth;
		float h = tileImage->imageHeight;
		
		tileImage->isTiled = true;
		tileImage->offsetX = phase.x();
		tileImage->offsetY =  phase.y();
		tileImage->srcWidth = tileImage->imageWidth;
		tileImage->srcHeight = tileImage->imageHeight;
		tileImage->destWidth = destRect.width();
		tileImage->destHeight = destRect.height();
		tileImage->x = destRect.x();
		tileImage->y = destRect.y();
		tileImage->tileWidth = tileRect.width();
		tileImage->tileHeight = tileRect.height();		
		tileImage->platform_data = getPlatformData();
		context->ag_backend->ag_draw_image(context, tileImage);
		context->ag_backend->ag_restore(context); 
	}


	void BitmapImage::draw(GraphicsContext* ctxt, const FloatRect& dstRect, const FloatRect& srcRect, CompositeOperator compositeOp)
	{
		if (!ctxt || !ctxt->platformContext() || !m_source.initialized())
			return;

#if HANDLE_PAGE_BREAKS
		if (ctxt->isDrawnOnLastPage(dstRect.y(), dstRect.height()))
			return;
#endif

		AGContext_t* context = ctxt->platformContext();

		AGImage_t* image = frameAtIndex(m_currentFrame);
		if (!image) // If it's too early we won't have an image yet.
			return;

		if (mayFillWithSolidColor()) {
			fillWithSolidColor(ctxt, dstRect, solidColor(), compositeOp);
			return;
		}

		image->isTiled = false;
		image->offsetX = srcRect.x();
		image->offsetY =  srcRect.y();
		image->srcWidth = srcRect.width();
		image->srcHeight = srcRect.height();
		image->destWidth = dstRect.width();
		image->destHeight = dstRect.height();
		image->x = dstRect.x();
		image->y = dstRect.y();
		image->platform_data = getPlatformData();
		context->ag_backend->ag_save(context);
		context->ag_backend->ag_draw_image(context, image);
		context->ag_backend->ag_restore(context);
	}
}
#endif //PLATFORM(AG)