//this include is necessary to build on mac.
//otherwise we get this compiler error: if you use new/delete make sure to include config.h at the top of the file
#include "config.h"
#include "ImageSource.h"
#include "GIFImageDecoder.h"
#include "JPEGImageDecoder.h"
#include "PNGImageDecoder.h"
#include "BMPImageDecoder.h"
#include "ICOImageDecoder.h"
#include "XBMImageDecoder.h"

#include "ag.h"
#include "agUtils.h"

namespace WebCore {

	ImageDecoder* createDecoder(SharedBuffer* buffer)
	{
		// We need at least 4 bytes to figure out what kind of image we're dealing with.
		int length = buffer->size();
		if (length < 4)
			return 0;

		const unsigned char* uContents = (const unsigned char*)buffer->data();
		const char* contents = buffer->data();

		// GIFs begin with GIF8(7 or 9).
		if (strncmp(contents, "GIF8", 4) == 0)
			return new GIFImageDecoder();

		// Test for PNG.
		if (uContents[0]==0x89 &&
			uContents[1]==0x50 &&
			uContents[2]==0x4E &&
			uContents[3]==0x47)
			return new PNGImageDecoder();

		// JPEG
		if (uContents[0]==0xFF &&
			uContents[1]==0xD8 &&
			uContents[2]==0xFF)
			return new JPEGImageDecoder();

		// BMP
		if (strncmp(contents, "BM", 2) == 0)
			return new BMPImageDecoder();

		// ICOs always begin with a 2-byte 0 followed by a 2-byte 1.
		// CURs begin with 2-byte 0 followed by 2-byte 2.
		if (!memcmp(contents, "\000\000\001\000", 4) ||
			!memcmp(contents, "\000\000\002\000", 4))
			return new ICOImageDecoder();

		// XBMs require 8 bytes of info.
		if (length >= 8 && strncmp(contents, "#define ", 8) == 0)
			return new XBMImageDecoder();

		// Give up. We don't know what the heck this is.
		return 0;
	}

	ImageSource::ImageSource()
		: m_decoder(0)
	{}

	ImageSource::~ImageSource()
	{
		delete m_decoder;
	}

	void ImageSource::clear()
	{
		if (m_decoder) {
			//CFRelease(m_decoder);
			m_decoder = 0;
		}
	}

	bool ImageSource::initialized() const
	{
		return m_decoder;
	}

	void ImageSource::setData(SharedBuffer* buffer, bool allDataReceived)
	{
		// Make the decoder by sniffing the bytes.
		// This method will examine the data and instantiate an instance of the appropriate decoder plugin.
		// If insufficient bytes are available to determine the image type, no decoder plugin will be
		// made.
		if(m_decoder || (m_decoder = createDecoder(buffer)) != NULL)
			m_decoder->setData(buffer, allDataReceived);
	}

	bool ImageSource::isSizeAvailable()
	{
		if (!m_decoder)
			return false;

		return m_decoder->isSizeAvailable();
	}

	IntSize ImageSource::size() const
	{
		if (!m_decoder)
			return IntSize();

		return m_decoder->size();
	}

	int ImageSource::repetitionCount()
	{
		if (!m_decoder)
			return cAnimationNone;

		return m_decoder->repetitionCount();
	}

	size_t ImageSource::frameCount() const
	{
		return m_decoder ? m_decoder->frameCount() : 0;
	}

	NativeImagePtr ImageSource::createFrameAtIndex(size_t index)
	{
		if (!m_decoder)
			return 0;

		RGBA32Buffer* buffer = m_decoder->frameBufferAtIndex(index);
		if (!buffer || buffer->status() == RGBA32Buffer::FrameEmpty)
			return 0;

		return ag_image_create((unsigned char*)buffer->bytes().data(), AG_IMAGE_FORMAT_ARGB32, 
							size().width(), 
							buffer->height(),
							size().width()*4, 32);
	}

	float ImageSource::frameDurationAtIndex(size_t index)
	{
		if (!m_decoder)
			return 0;

		RGBA32Buffer* buffer = m_decoder->frameBufferAtIndex(index);
		if (!buffer || buffer->status() == RGBA32Buffer::FrameEmpty)
			return 0;

		return buffer->duration() / 1000.0f;
	}

	bool ImageSource::frameHasAlphaAtIndex(size_t index)
	{
		// Might be interesting to do this optimization on Mac some day, but for now we're just using this
		// for the Cairo source, since it uses our decoders, and our decoders can answer this question.
		// FIXME: Could return false for JPEG and other non-transparent image formats.
		// FIXME: Could maybe return false for a GIF Frame if we have enough info in the GIF properties dictionary
		// to determine whether or not a transparent color was defined.
		return true;
	}

	bool ImageSource::frameIsCompleteAtIndex(size_t index)
	{
		assert(false);
		return false;
	}

}
