/*************************************************************************
*
* ADOBE CONFIDENTIAL
* ___________________
*
*  Copyright 2007 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains
* the property of Adobe Systems Incorporated and its suppliers,
* if any.  The intellectual and technical concepts contained
* herein are proprietary to Adobe Systems Incorporated and its
* suppliers and may be covered by U.S. and Foreign Patents,
* patents in process, and are protected by trade secret or copyright law.
* Dissemination of this information or reproduction of this material
* is strictly forbidden unless prior written permission is obtained
* from Adobe Systems Incorporated.
**************************************************************************/
#include "config.h"

#if PLATFORM(AG)

#include "ag.h"
#include "agUtils.h"

#if PLATFORM(WIN)
#include "malloc.h"
#else
#include <sys/malloc.h>
#endif

void ag_matrix_multiply (AGMatrix_t *result, const AGMatrix_t *m1, const AGMatrix_t *m2)
{
    AGMatrix_t r;

	r.a = m1->a * m2->a + m1->b * m2->c;
	r.b = m1->a * m2->b + m1->b * m2->d;
	r.c = m1->c * m2->a + m1->d * m2->c;
	r.d = m1->c * m2->b + m1->d * m2->d;

	r.h = m1->h * m2->a + m1->v * m2->c + m2->h;
	r.v = m1->h * m2->b + m1->v * m2->d + m2->v;

    *result = r;
}

void ag_matrix_transform_distance (const AGMatrix_t *matrix, double *dx, double *dy)
{
    double new_x, new_y;

    new_x = (matrix->a * *dx + matrix->c * *dy);
    new_y = (matrix->b * *dx + matrix->d * *dy);

    *dx = new_x;
    *dy = new_y;
}


void ag_user_to_device (AGContext_t *cr, double *x, double *y)
{
	AGMatrix_t m;
	cr->ag_backend->ag_get_CTM(cr, &m);
	ag_matrix_transform_distance(&m, x, y);
	*x += m.h;
	*y += m.v;
}

void ag_device_to_user (AGContext_t *cr, double *x, double *y)
{
}

void ag_user_to_device_distance (AGContext_t *cr, double *x, double *y)
{
	AGMatrix_t m;
	cr->ag_backend->ag_get_CTM(cr, &m);
	ag_matrix_transform_distance(&m, x, y);
}

void ag_device_to_user_distance (AGContext_t *cr, double *dx, double *dy)
{
}

void ag_translate(AGContext_t *cr, double dx, double dy)
{
	AGMatrix_t m, m2, r;
	ag_init_matrix_to_identity(&m);
	cr->ag_backend->ag_get_CTM(cr, &m);
	m2.a = 1; m2.b = 0; m2.c = 0; m2.d = 1; m2.h = dx; m2.v = dy;
	ag_matrix_multiply(&r, &m2, &m);
	cr->ag_backend->ag_set_CTM(cr, &r);
}

void ag_scale(AGContext_t *cr, double dx, double dy)
{
	assert(false);
}

void ag_rotate(AGContext_t *cr, double angle)
{
	assert(false);
}

void ag_init_matrix_to_identity(AGMatrix_t* m)
{
	m->a = 1; m->b = 0; m->c = 0; m->d = 1; m->h = 0; m->v = 0;
}

void ag_matrix_init_scale(AGMatrix_t* m, double sx, double sy)
{
	AGMatrix_t m2 = { sx, 0, 0, sy, 0, 0};
	ag_matrix_multiply(m, m, &m2);
}

void ACCB2 ImageMemBufDeleteCallBack(void * p) {
	// In the current impl, the image data is destroyed via FrameData::clear()
}

AGImage_t* ag_image_create(unsigned char *data, AGImageFormat_t	format, int width, int height, int stride, int depth)
{
	assert(format == AG_IMAGE_FORMAT_ARGB32);
	assert(stride == width * 4);
	assert(reinterpret_cast<unsigned int>(data) % 4 == 0); // Data must be alligned

	AGImage_t* agImage = (AGImage_t*)malloc(sizeof(AGImage_t));

	if (agImage)
	{
		memset(agImage, 0, sizeof(AGImage_t));

		agImage->hasAlpha = false;
		unsigned char *ptr = (unsigned char *)data;
		for (int j = 0; j < height*width*4; j+=4) 
		{
			if(ptr[j] != 0xff)
			{
				agImage->hasAlpha = true;
				break;
			}
		}

		int rgbaSize = agImage->hasAlpha ? height * width * 4 : height * width * 3;
		
		unsigned char* rgba = (unsigned char *)malloc(rgbaSize);

		if (!rgba) {
			return 0;
		}

		unsigned char * pDest = rgba, * pSrcBackwards = data + (height * stride);
		
		if(agImage->hasAlpha)
		{
			for (int i = 0; i < height; ++i) {
				pSrcBackwards -= stride;
				memcpy(pDest, pSrcBackwards, stride);
				pDest += stride;
			}
		}
		else
		{
			int pDestStride = width * 3;

			for (int i = 0; i < height; ++i) {
				pSrcBackwards -= stride;
				
				for(int k = 0; k < width; ++k)
				{
					pDest[k*3] = pSrcBackwards[k*4 + 1];
					pDest[k*3+1] = pSrcBackwards[k*4 + 2];
					pDest[k*3+2] = pSrcBackwards[k*4 + 3];
				}

				pDest += pDestStride;
			}
		}		

		agImage->rgbaData = rgba;
		agImage->rgbaDataSize = rgbaSize;
		agImage->format = format;
		agImage->imageHeight = height;
		agImage->imageWidth = width;
		agImage->depth = depth;
		agImage->platform_data = 0;
		agImage->image_del_cb_func = ImageMemBufDeleteCallBack;
	}
	return agImage;
}

void ag_image_destroy(AGImage_t* agImage)
{
	if (agImage->rgbaData)
		free(agImage->rgbaData);
	free(agImage);
}

#endif //PLATFORM(AG)